package ru.dbotthepony.mc.otm.entity

import net.minecraft.core.particles.ParticleTypes
import net.minecraft.network.syncher.EntityDataAccessor
import net.minecraft.network.syncher.EntityDataSerializers
import net.minecraft.network.syncher.SynchedEntityData
import net.minecraft.server.level.ServerBossEvent
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.sounds.SoundEvent
import net.minecraft.sounds.SoundEvents
import net.minecraft.sounds.SoundSource
import net.minecraft.util.Mth
import net.minecraft.world.BossEvent
import net.minecraft.world.damagesource.DamageSource
import net.minecraft.world.entity.AnimationState
import net.minecraft.world.entity.EntityType
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.ai.attributes.AttributeSupplier
import net.minecraft.world.entity.ai.attributes.Attributes
import net.minecraft.world.entity.ai.goal.Goal
import net.minecraft.world.entity.ai.goal.LookAtPlayerGoal
import net.minecraft.world.entity.ai.goal.target.HurtByTargetGoal
import net.minecraft.world.entity.ai.goal.target.NearestAttackableTargetGoal
import net.minecraft.world.entity.ai.navigation.GroundPathNavigation
import net.minecraft.world.entity.ai.navigation.PathNavigation
import net.minecraft.world.entity.monster.AbstractIllager
import net.minecraft.world.entity.monster.AbstractSkeleton
import net.minecraft.world.entity.monster.Monster
import net.minecraft.world.entity.monster.Zombie
import net.minecraft.world.entity.npc.Villager
import net.minecraft.world.entity.player.Player
import net.minecraft.world.entity.projectile.SmallFireball
import net.minecraft.world.level.Level
import net.minecraft.world.phys.Vec3
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.registry.game.MSoundEvents
import java.util.*

class Enforcer(type: EntityType<Enforcer>, level: Level) : Monster(type,level) {
	val idleState = AnimationState()
	val chargeState = AnimationState()

	init {
		idleState.start(tickCount)
	}

	companion object {
		fun createAttributes() : AttributeSupplier.Builder {
			return createMonsterAttributes()
				.add(Attributes.MAX_HEALTH, 300.0)
				.add(Attributes.ARMOR, 20.0)
				.add(Attributes.MOVEMENT_SPEED, 0.3)
				.add(Attributes.STEP_HEIGHT, 1.0)
				.add(Attributes.KNOCKBACK_RESISTANCE, 1.0)
				.add(Attributes.FOLLOW_RANGE, 32.0)
		}
		private val IS_CHARGING: EntityDataAccessor<Boolean> =
			SynchedEntityData.defineId(Enforcer::class.java, EntityDataSerializers.BOOLEAN)
	}

	override fun defineSynchedData(builder: SynchedEntityData.Builder) {
		super.defineSynchedData(builder)
		builder.define(IS_CHARGING, false)
	}

	fun setCharging(value: Boolean) {
		entityData.set(IS_CHARGING, value)
	}

	fun isCharging(): Boolean {
		return entityData.get(IS_CHARGING)
	}

	override fun registerGoals() {
		goalSelector.addGoal(7, LookAtPlayerGoal(this, Player::class.java, 8f))
		goalSelector.addGoal(3, NearestAttackableTargetGoal(this, LivingEntity::class.java, 10, true, true) { entity ->
			entity is Player ||
				entity is Villager ||
				entity is AbstractIllager ||
				entity is Zombie ||
				entity is AbstractSkeleton
		})

		goalSelector.addGoal(2, RammingGoal(this))
		goalSelector.addGoal(2, StayNearGoal(this))
		goalSelector.addGoal(2, BlazeFireballGoal(this))

		targetSelector.addGoal(1, HurtByTargetGoal(this))
	}

	override fun createNavigation(level: Level): PathNavigation = GroundPathNavigation(this, level)

	override fun isPushable(): Boolean {
		return false
	}

	override fun getHurtSound(damageSource: DamageSource): SoundEvent {
		return SoundEvents.HEAVY_CORE_BREAK
	}

	override fun getDeathSound(): SoundEvent {
		return SoundEvents.VAULT_BREAK
	}

	//fire inv
	//there should be a better way
	override fun fireImmune(): Boolean {
		return true
	}

	///boss healthbar
	private val bossEvent: ServerBossEvent =
		ServerBossEvent(TranslatableComponent(MNames.ENFORCER), BossEvent.BossBarColor.RED, BossEvent.BossBarOverlay.PROGRESS)

	override fun startSeenByPlayer(player: ServerPlayer) {
		super.startSeenByPlayer(player)
		bossEvent.addPlayer(player)
	}

	override fun stopSeenByPlayer(player: ServerPlayer) {
		super.stopSeenByPlayer(player)
		bossEvent.removePlayer(player)
	}

	override fun aiStep() {
		super.aiStep()
		bossEvent.progress = this.health / this.maxHealth

		if (!this.isOnFire) {
			this.clearFire()
		}
	}

	override fun die(cause: DamageSource) {
		super.die(cause)
		bossEvent.removeAllPlayers()
	}
	///

	//this stuff removes 180-degree head rotation limits that minecraft mobs have, since this is a turret on a tank, we don't need such limits
	//WE ARE BEYOND THAT, THOSE ARE MY GREAT WAR-MACHINES
	override fun getMaxHeadYRot(): Int {
		return 180
	}

	fun shootFireball(math: Float) {
		if (level().isClientSide) return

		val offset_dist = 1.4
		val viewVector = this.getViewVector(1.0F).normalize()
		val leftVec = Vec3(viewVector.z, 0.0, -viewVector.x).normalize().scale(offset_dist)

		val gun_x = x + leftVec.x
		val gun_z = z + leftVec.z
		val gun_y = y + 1.2

		//stfu idea i like my underscores

		val fireballX = gun_x
		val fireballY = gun_y
		val fireballZ = gun_z

		val fireball = SmallFireball(level(), fireballX, fireballY, fireballZ, viewVector)
		fireball.shootFromRotation(this, this.xRot, this.yHeadRot, 0.0F, 1.5F, 0.2F)
		fireball.setPos(fireballX, fireballY, fireballZ)
		fireball.owner = this

		level().addFreshEntity(fireball)
	}

	fun shootMissile(math: Float) {
		if (level().isClientSide) return

		val offset_dist = 1.4
		val viewVector = this.getViewVector(1.0F).normalize()
		val leftVec = Vec3(-viewVector.z, 0.0, viewVector.x).normalize().scale(offset_dist)

		val gun_x = x + leftVec.x
		val gun_z = z + leftVec.z
		val gun_y = y + 1.2

		//stfu idea i like my underscores

		val missileX = gun_x
		val missileY = gun_y
		val missileZ = gun_z

		val fireball = SmallFireball(level(), missileX, missileY, missileZ, viewVector)
		fireball.shootFromRotation(this, this.xRot, this.yHeadRot, 0.0F, 1.5F, 0.2F)
		fireball.setPos(missileX, missileY, missileZ)
		fireball.owner = this

		level().addFreshEntity(fireball)
	}

	//charge attack
	class RammingGoal(private val mob: Enforcer) : Goal() {
		private var target: LivingEntity? = null

		init {
			setFlags(EnumSet.of(Goal.Flag.MOVE))
		}

		private var chargeTime = 0
		private val windupTime = 20
		private val maxChargeTime = 40
		private var chargeDir: Vec3? = null
		private var cooldown = 0
		private val minCooldown = 35
		private val maxCooldown = 60
		private var isCharging = false

		override fun canUse(): Boolean {
			if (cooldown > 0) {
				cooldown--
				return false
			}

			target = mob.target
			return target != null
		}

		override fun canContinueToUse(): Boolean {
			return isCharging && chargeTime < maxChargeTime
		}

		override fun start() {
			mob.setCharging(true)
			mob.playSound(MSoundEvents.ENFORCER_ALERT, 1.0f, 1.0f)
			chargeTime = 0
			mob.navigation.stop()
			cooldown = minCooldown + mob.random.nextInt(maxCooldown - minCooldown)
			isCharging = true
			mob.chargeState.start(mob.tickCount)
		}

		override fun tick() {
			if (!isCharging) return

			chargeTime++
			val target = mob.target ?: return

			if (chargeTime < windupTime) {
				chargeDir = Vec3(target.x - mob.x, 0.0, target.z - mob.z).normalize()

				mob.lookControl.setLookAt(target, 30.0f, 30.0f)
				return
			}

			val leftPartic_x = mob.x - mob.lookAngle.x * 2 - 0.8f * mob.lookAngle.z
			val rightPartic_x = mob.x - mob.lookAngle.x * 2 + 0.8f * mob.lookAngle.z
			val smoke_y = mob.y + 0.6f
			val leftPartic_z = mob.z - mob.lookAngle.z * 2 + 0.8f * mob.lookAngle.x
			val rightPartic_z = mob.z - mob.lookAngle.z * 2 - 0.8f * mob.lookAngle.x

			if (mob.level() is ServerLevel) {
				(mob.level() as ServerLevel).sendParticles(
					ParticleTypes.LARGE_SMOKE,
					leftPartic_x, smoke_y, leftPartic_z,
					2,
					0.0, 0.07, 0.0,
					0.01
				)

				(mob.level() as ServerLevel).sendParticles(
					ParticleTypes.LARGE_SMOKE,
					rightPartic_x, smoke_y, rightPartic_z,
					2,
					0.0, 0.07, 0.0,
					0.01
				)
			}

			if (chargeTime == windupTime) {
				if (chargeDir == null || chargeDir!!.length() < 0.1) {
					stop()
					return
				}
				mob.playSound(MSoundEvents.ENFORCER_CHARGE, 1.0f, 1.0f)
			}

			val dir = chargeDir ?: return
			mob.yRot = (-Mth.atan2(dir.x, dir.z) * (180f / Math.PI)).toFloat()
			mob.yHeadRot = mob.yRot

			mob.move(net.minecraft.world.entity.MoverType.SELF, dir.scale(1.8))


			if (mob.horizontalCollision) {
				mob.level().playSound(
					null,
					mob.x, mob.y, mob.z,
					SoundEvents.GENERIC_EXPLODE,
					SoundSource.HOSTILE,
					1.0f,
					1.0f
				)
				stop()
				return
			}

			if (mob.boundingBox.intersects(target.boundingBox)) {
				val knockbackStrength = 1.5
				val knockback = dir.scale(knockbackStrength)

				target.push(knockback.x, 0.5, knockback.z)

				val damageAmount = 14.0f
				target.hurt(mob.damageSources().mobAttack(mob), damageAmount)

				stop()
			}
		}

		override fun stop() {
			mob.setCharging(false)
			mob.setDeltaMovement(Vec3.ZERO)
			chargeTime = 0
			chargeDir = null
			isCharging = false
			mob.chargeState.stop()
		}
	}

	class StayNearGoal(private val mob: Enforcer) : Goal() {
		private var target: LivingEntity? = null
		private var moveCD = 0
		private var movePos: Vec3? = null

		override fun canUse(): Boolean {
			return mob.target != null && !mob.isCharging()
		}

		override fun canContinueToUse(): Boolean {
			return canUse()
		}

		override fun start() {
			target = mob.target
			moveCD = 0
		}

		override fun tick() {
			mob.lookControl.setLookAt(target, 30.0f, 30.0f)
			val target = mob.target ?: return

			if (moveCD > 0) {
				moveCD--
				return
			}

			moveCD = 10 + mob.random.nextInt(10)

			val targetX = target.x
			val targetZ = target.z

			val angle = mob.random.nextDouble() * Math.PI * 2
			val distance = 8.0 + mob.random.nextDouble() * 3.0

			val offsetX = Math.cos(angle) * distance
			val offsetZ = Math.sin(angle) * distance

			movePos = Vec3(targetX + offsetX, target.y + 2.0, targetZ + offsetZ)

			mob.navigation.moveTo(movePos!!.x, movePos!!.y, movePos!!.z, 1.5)
		}
	}

	class BlazeFireballGoal(private val mob: Enforcer) : Goal() {
		private var cooldown = 0
		private var fireTick = 0

		init {
			setFlags(EnumSet.of(Goal.Flag.LOOK))
		}

		override fun canUse(): Boolean {
			return mob.target != null && !mob.isCharging()
		}

		override fun canContinueToUse(): Boolean {
			return canUse()
		}

		override fun tick() {
			val target = mob.target ?: return
			mob.lookControl.setLookAt(target, 30.0f, 30.0f)

			if (cooldown > 0) {
				fireTick = 20
				cooldown--
				return
			}

			if (fireTick > 0){
				fireTick --


				if (fireTick >= 19){
					mob.playSound(MSoundEvents.ENFORCER_BEEP, 1.0f, 1.0f)
				}


				if (fireTick == 15){
					mob.shootFireball(0.5f)
					//could have a better firing sound
					mob.playSound(SoundEvents.BLAZE_SHOOT, 1.0f, 1.0f)
				}
				if (fireTick == 10){
					mob.shootFireball(0.5f)
					mob.playSound(SoundEvents.BLAZE_SHOOT, 1.0f, 1.0f)
				}
				if (fireTick == 5){
					mob.shootFireball(0.5f)
					mob.playSound(SoundEvents.BLAZE_SHOOT, 1.0f, 1.0f)
				}

				return
			} else {
				cooldown = 5 + mob.random.nextInt(10)
			}
		}
	}
}
